#define _POSIX_C_SOURCE 199309L
#include <time.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include "simd-1.h"
const double processor_ghz = 5.5;


double* random_double_array(uint64_t length) {
    double* data = (double*)malloc(length * sizeof(double));
    for (uint64_t i = 0; i < length; i++) {
        data[i] = (double)rand() / (double)RAND_MAX - 0.2;
    }
    return data;
}

float* to_single(double* doubles, uint64_t length) {
    float* floats = (float*)malloc(length * sizeof(float));
    if (floats == NULL) {
        return NULL;
    }
    for (uint64_t i = 0; i < length; i++) {
        floats[i] = (float)doubles[i];
    }
    return floats;
}

void test_float_sum(char* descr, float float_sum(float *, uint64_t), float* data, uint64_t length) {
    struct timespec start, end;
    float total;
    clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &start);
    total = float_sum(data, length);
    clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &end);
    double elapsed = (end.tv_sec - start.tv_sec) + (end.tv_nsec - start.tv_nsec) / 1e9;
    printf("%s: calculated %.10g in %g s, %.1f cycles/element.\n", descr, total, elapsed, elapsed * 1e9 / length * processor_ghz);
}
void test_double_sum(char* descr, double double_sum(double *, uint64_t), double* data, uint64_t length) {
    struct timespec start, end;
    double total;
    clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &start);
    total = double_sum(data, length);
    clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &end);
    double elapsed = (end.tv_sec - start.tv_sec) + (end.tv_nsec - start.tv_nsec) / 1e9;
    printf("%s: calculated %.10g in %g s, %.1f cycles/element.\n", descr, total, elapsed, elapsed * 1e9 / length * processor_ghz);
}

float float_sum_c(float* data, uint64_t length) {
    float total = 0;
    for (uint64_t i = 0; i < length; i++) {
        total += data[i];
    }
    return total;
}

double double_sum_c(double* data, uint64_t length) {
    double total = 0;
    for (uint64_t i = 0; i < length; i++) {
        total += data[i];
    }
    return total;
}

int main(int argc, char* argv[]) {
    srand(time(NULL));
    if (argc < 2) {
        printf("Must give length on command line.\n");
        return 1;
    }
    uint64_t length = atol(argv[1]);
    if ( length % 8 != 0 ) {
        printf("Length must be divisible by 8.\n");
        return 1;
    }
    
    double small[16] = {0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8,  0.9, 1.0,  1.1, 1.2, 1.3, 1.4, 1.5, 1.6};
    float* smallf = to_single(small, 16);
    
    printf("sC %g\n", float_sum_c(smallf, 16));
    printf("s1 %g\n", float_sum_1(smallf, 16));
    printf("s2 %g\n", float_sum_2(smallf, 16));
    printf("dC %g\n", double_sum_c(small, 16));
    printf("d1 %g\n", double_sum_1(small, 16));
    printf("d2 %g\n", double_sum_2(small, 16));
    
    double* doubles = random_double_array(length);
    float* floats = to_single(doubles, length);
    printf("Array sizes: %luMB and %luMB.\n", sizeof(double)*length/1024/1024, sizeof(float)*length/1024/1024);
    
    test_float_sum("sC", float_sum_c, floats, length);
    test_float_sum("s1", float_sum_1, floats, length);
    test_float_sum("s2", float_sum_2, floats, length);
    test_double_sum("dC", double_sum_c, doubles, length);
    test_double_sum("d1", double_sum_1, doubles, length);
    test_double_sum("d2", double_sum_2, doubles, length);
    
    free(smallf);
    free(floats);
    free(doubles);
    return 0;
}
